# -*- coding: utf-8 -*-

"""Collection of application detection components."""

from .core import *
from .filters import *
from .folders import *
from .groups import *
from .sources import *
from .utilities import *
